// Copyright (c) 2023, INRAE.
// Distributed under the terms of the GPL-3 Licence.
// The full licence is in the file LICENCE, distributed with this software.

#ifndef EVALHYD_DETERMINIST_ELEMENTS_HPP
#define EVALHYD_DETERMINIST_ELEMENTS_HPP

#include <xtensor/xtensor.hpp>
#include <xtensor/xview.hpp>

#include "../maths.hpp"


namespace evalhyd
{
    namespace determinist
    {
        namespace elements
        {
            /// Compute the mean of the observations.
            ///
            /// \param q_obs
            ///     Streamflow observations.
            ///     shape: (1, time)
            /// \param t_msk
            ///     Temporal subsets of the whole record.
            ///     shape: (series, subsets, time)
            /// \param b_exp
            ///     Boostrap samples.
            ///     shape: (samples, time slice)
            /// \param n_srs
            ///     Number of prediction series.
            /// \param n_msk
            ///     Number of temporal subsets.
            /// \param n_exp
            ///     Number of bootstrap samples.
            /// \return
            ///     Mean observed streamflow.
            ///     shape: (subsets, samples, series, 1)
            template <class XD2o>
            inline xt::xtensor<double, 4> calc_mean_obs(
                    const XD2o& q_obs,
                    const xt::xtensor<bool, 3>& t_msk,
                    const std::vector<xt::xkeep_slice<int>>& b_exp,
                    std::size_t n_srs,
                    std::size_t n_msk,
                    std::size_t n_exp
            )
            {
                xt::xtensor<double, 4> mean_obs =
                        xt::zeros<double>({n_msk, n_exp, n_srs, std::size_t {1}});

                for (std::size_t m = 0; m < n_msk; m++)
                {
                    // apply the mask
                    // (using NaN workaround until reducers work on masked_view)
                    auto obs_masked = xt::where(xt::view(t_msk, xt::all(), m),
                                                q_obs, NAN);

                    // compute variable one sample at a time
                    for (std::size_t e = 0; e < n_exp; e++)
                    {
                        // apply the bootstrap sampling
                        auto obs = xt::view(obs_masked, xt::all(), b_exp[e]);
                        xt::view(mean_obs, m, e) =
                                xt::nanmean(obs, -1, xt::keep_dims);
                    }
                }

                return mean_obs;
            }

            /// Compute the mean of the predictions.
            ///
            /// \param q_prd
            ///     Streamflow predictions.
            ///     shape: (series, time)
            /// \param t_msk
            ///     Temporal subsets of the whole record.
            ///     shape: (series, subsets, time)
            /// \param b_exp
            ///     Boostrap samples.
            ///     shape: (samples, time slice)
            /// \param n_srs
            ///     Number of prediction series.
            /// \param n_msk
            ///     Number of temporal subsets.
            /// \param n_exp
            ///     Number of bootstrap samples.
            /// \return
            ///     Mean predicted streamflow.
            ///     shape: (subsets, samples, series, 1)
            template <class XD2p>
            inline xt::xtensor<double, 4> calc_mean_prd(
                    const XD2p& q_prd,
                    const xt::xtensor<bool, 3>& t_msk,
                    const std::vector<xt::xkeep_slice<int>>& b_exp,
                    std::size_t n_srs,
                    std::size_t n_msk,
                    std::size_t n_exp
            )
            {
                xt::xtensor<double, 4> mean_prd =
                        xt::zeros<double>({n_msk, n_exp, n_srs, std::size_t {1}});

                for (std::size_t m = 0; m < n_msk; m++)
                {
                    // apply the mask
                    // (using NaN workaround until reducers work on masked_view)
                    auto prd_masked = xt::where(xt::view(t_msk, xt::all(), m),
                                                q_prd, NAN);

                    // compute variable one sample at a time
                    for (std::size_t e = 0; e < n_exp; e++)
                    {
                        // apply the bootstrap sampling
                        auto prd = xt::view(prd_masked, xt::all(), b_exp[e]);
                        xt::view(mean_prd, m, e) =
                                xt::nanmean(prd, -1, xt::keep_dims);
                    }
                }

                return mean_prd;
            }

            /// Compute the quadratic error between observations and predictions.
            ///
            /// \param q_obs
            ///     Streamflow observations.
            ///     shape: (1, time)
            /// \param q_prd
            ///     Streamflow predictions.
            ///     shape: (series, time)
            /// \return
            ///     Quadratic errors between observations and predictions.
            ///     shape: (series, time)
            template <class XD2o, class XD2p>
            inline xt::xtensor<double, 2> calc_quad_err(
                    const XD2o& q_obs,
                    const XD2p& q_prd
            )
            {
                return xt::square(q_obs - q_prd);
            }

            /// Compute the quadratic error between observations and mean observation.
            ///
            /// \param q_obs
            ///     Streamflow observations.
            ///     shape: (series, time)
            /// \param mean_obs
            ///     Mean observed streamflow.
            ///     shape: (subsets, samples, series, 1)
            /// \param t_msk
            ///     Temporal subsets of the whole record.
            ///     shape: (series, subsets, time)
            /// \param n_srs
            ///     Number of prediction series.
            /// \param n_tim
            ///     Number of time steps.
            /// \param n_msk
            ///     Number of temporal subsets.
            /// \param n_exp
            ///     Number of bootstrap samples.
            /// \return
            ///     Quadratic errors between observations and mean observation.
            ///     shape: (subsets, samples, series, time)
            template <class XD2o>
            inline xt::xtensor<double, 4> calc_quad_obs(
                    const XD2o& q_obs,
                    const xt::xtensor<double, 4>& mean_obs,
                    const xt::xtensor<bool, 3>& t_msk,
                    std::size_t n_srs,
                    std::size_t n_tim,
                    std::size_t n_msk,
                    std::size_t n_exp
            )
            {
                xt::xtensor<double, 4> quad_obs =
                        xt::zeros<double>({n_msk, n_exp, n_srs, n_tim});

                for (std::size_t m = 0; m < n_msk; m++)
                {
                    // apply the mask
                    // (using NaN workaround until reducers work on masked_view)
                    auto obs_masked = xt::where(xt::view(t_msk, xt::all(), m),
                                                q_obs, NAN);

                    for (std::size_t e = 0; e < n_exp; e++)
                    {
                        xt::view(quad_obs, m, e) = xt::square(
                                obs_masked - xt::view(mean_obs, m, e)
                        );
                    }
                }

                return quad_obs;
            }

            /// Compute the quadratic error between predictions and mean prediction.
            ///
            /// \param q_prd
            ///     Streamflow predictions.
            ///     shape: (series, time)
            /// \param mean_prd
            ///     Mean predicted streamflow.
            ///     shape: (subsets, samples, series, 1)
            /// \param t_msk
            ///     Temporal subsets of the whole record.
            ///     shape: (series, subsets, time)
            /// \param n_srs
            ///     Number of prediction series.
            /// \param n_tim
            ///     Number of time steps.
            /// \param n_msk
            ///     Number of temporal subsets.
            /// \param n_exp
            ///     Number of bootstrap samples.
            /// \return
            ///     Quadratic errors between predictions and mean prediction.
            ///     shape: (subsets, samples, series, time)
            template <class XD2p>
            inline xt::xtensor<double, 4> calc_quad_prd(
                    const XD2p& q_prd,
                    const xt::xtensor<double, 4>& mean_prd,
                    const xt::xtensor<bool, 3>& t_msk,
                    std::size_t n_srs,
                    std::size_t n_tim,
                    std::size_t n_msk,
                    std::size_t n_exp
            )
            {
                xt::xtensor<double, 4> quad_prd =
                        xt::zeros<double>({n_msk, n_exp, n_srs, n_tim});

                for (std::size_t m = 0; m < n_msk; m++)
                {
                    // apply the mask
                    // (using NaN workaround until reducers work on masked_view)
                    auto prd_masked = xt::where(xt::view(t_msk, xt::all(), m),
                                                q_prd, NAN);

                    for (std::size_t e = 0; e < n_exp; e++)
                    {
                        xt::view(quad_prd, m, e) = xt::square(
                                prd_masked - xt::view(mean_prd, m, e)
                        );
                    }
                }

                return quad_prd;
            }

            /// Compute the Pearson correlation coefficient.
            ///
            /// \param q_obs
            ///     Streamflow observations.
            ///     shape: (series, time)
            /// \param q_prd
            ///     Streamflow predictions.
            ///     shape: (series, time)
            /// \param mean_obs
            ///     Mean observed streamflow.
            ///     shape: (subsets, samples, series, 1)
            /// \param mean_prd
            ///     Mean predicted streamflow.
            ///     shape: (subsets, samples, series, 1)
            /// \param quad_obs
            ///     Quadratic errors between observations and mean observation.
            ///     shape: (subsets, samples, series, time)
            /// \param quad_prd
            ///     Quadratic errors between predictions and mean prediction.
            ///     shape: (subsets, samples, series, time)
            /// \param t_msk
            ///     Temporal subsets of the whole record.
            ///     shape: (series, subsets, time)
            /// \param b_exp
            ///     Boostrap samples.
            ///     shape: (samples, time slice)
            /// \param n_srs
            ///     Number of prediction series.
            /// \param n_msk
            ///     Number of temporal subsets.
            /// \param n_exp
            ///     Number of bootstrap samples.
            /// \return
            ///     Pearson correlation coefficients.
            ///     shape: (subsets, samples, series)
            template <class XD2o, class XD2p>
            inline xt::xtensor<double, 3> calc_r_pearson(
                    const XD2o& q_obs,
                    const XD2p& q_prd,
                    const xt::xtensor<double, 4>& mean_obs,
                    const xt::xtensor<double, 4>& mean_prd,
                    const xt::xtensor<double, 4>& quad_obs,
                    const xt::xtensor<double, 4>& quad_prd,
                    const xt::xtensor<bool, 3>& t_msk,
                    const std::vector<xt::xkeep_slice<int>>& b_exp,
                    std::size_t n_srs,
                    std::size_t n_msk,
                    std::size_t n_exp
            )
            {
                // calculate error in timing and dynamics $r_{pearson}$
                // (Pearson's correlation coefficient)
                xt::xtensor<double, 3> r_pearson =
                        xt::zeros<double>({n_msk, n_exp, n_srs});

                for (std::size_t m = 0; m < n_msk; m++)
                {
                    // apply the mask
                    // (using NaN workaround until reducers work on masked_view)
                    auto prd_masked = xt::where(xt::view(t_msk, xt::all(), m),
                                                q_prd, NAN);
                    auto obs_masked = xt::where(xt::view(t_msk, xt::all(), m),
                                                q_obs, NAN);

                    // compute variable one sample at a time
                    for (std::size_t e = 0; e < n_exp; e++)
                    {
                        auto prd = xt::view(prd_masked, xt::all(), b_exp[e]);
                        auto obs = xt::view(obs_masked, xt::all(), b_exp[e]);
                        auto r_num = xt::nansum(
                                (prd - xt::view(mean_prd, m, e))
                                * (obs - xt::view(mean_obs, m, e)),
                                -1
                        );

                        auto prd2 = xt::view(quad_prd, m, e, xt::all(), b_exp[e]);
                        auto obs2 = xt::view(quad_obs, m, e, xt::all(), b_exp[e]);
                        auto r_den = xt::sqrt(
                                xt::nansum(prd2, -1) * xt::nansum(obs2, -1)
                        );

                        xt::view(r_pearson, m, e) = r_num / r_den;
                    }
                }

                return r_pearson;
            }

            /// Compute alpha.
            ///
            /// \param q_obs
            ///     Streamflow observations.
            ///     shape: (series, time)
            /// \param q_prd
            ///     Streamflow predictions.
            ///     shape: (series, time)
            /// \param mean_obs
            ///     Mean observed streamflow.
            ///     shape: (subsets, samples, series, 1)
            /// \param mean_prd
            ///     Mean predicted streamflow.
            ///     shape: (subsets, samples, series, 1)
            /// \param t_msk
            ///     Temporal subsets of the whole record.
            ///     shape: (series, subsets, time)
            /// \param b_exp
            ///     Boostrap samples.
            ///     shape: (samples, time slice)
            /// \param n_srs
            ///     Number of prediction series.
            /// \param n_msk
            ///     Number of temporal subsets.
            /// \param n_exp
            ///     Number of bootstrap samples.
            /// \return
            ///     Alphas, ratios of standard deviations.
            ///     shape: (subsets, samples, series)
            template <class XD2o, class XD2p>
            inline xt::xtensor<double, 3> calc_alpha(
                    const XD2o& q_obs,
                    const XD2p& q_prd,
                    const xt::xtensor<double, 4>& mean_obs,
                    const xt::xtensor<double, 4>& mean_prd,
                    const xt::xtensor<bool, 3>& t_msk,
                    const std::vector<xt::xkeep_slice<int>>& b_exp,
                    std::size_t n_srs,
                    std::size_t n_msk,
                    std::size_t n_exp
            )
            {
                // calculate error in spread of flow $alpha$
                xt::xtensor<double, 3> alpha =
                        xt::zeros<double>({n_msk, n_exp, n_srs});

                for (std::size_t m = 0; m < n_msk; m++)
                {
                    // apply the mask
                    // (using NaN workaround until reducers work on masked_view)
                    auto prd_masked = xt::where(xt::view(t_msk, xt::all(), m),
                                                q_prd, NAN);
                    auto obs_masked = xt::where(xt::view(t_msk, xt::all(), m),
                                                q_obs, NAN);

                    // compute variable one sample at a time
                    for (std::size_t e = 0; e < n_exp; e++)
                    {
                        auto prd = xt::view(prd_masked, xt::all(), b_exp[e]);
                        auto obs = xt::view(obs_masked, xt::all(), b_exp[e]);
                        xt::view(alpha, m, e) =
                                maths::nanstd(prd, xt::view(mean_prd, m, e))
                                / maths::nanstd(obs, xt::view(mean_obs, m, e));
                    }
                }

                return alpha;
            }

            /// Compute the bias.
            ///
            /// \param q_obs
            ///     Streamflow observations.
            ///     shape: (series, time)
            /// \param q_prd
            ///     Streamflow predictions.
            ///     shape: (series, time)
            /// \param t_msk
            ///     Temporal subsets of the whole record.
            ///     shape: (series, subsets, time)
            /// \param b_exp
            ///     Boostrap samples.
            ///     shape: (samples, time slice)
            /// \param n_srs
            ///     Number of prediction series.
            /// \param n_msk
            ///     Number of temporal subsets.
            /// \param n_exp
            ///     Number of bootstrap samples.
            /// \return
            ///     Biases.
            ///     shape: (subsets, samples, series)
            template <class XD2o, class XD2p>
            inline xt::xtensor<double, 3> calc_bias(
                    const XD2o& q_obs,
                    const XD2p& q_prd,
                    const xt::xtensor<bool, 3>& t_msk,
                    const std::vector<xt::xkeep_slice<int>>& b_exp,
                    std::size_t n_srs,
                    std::size_t n_msk,
                    std::size_t n_exp
            )
            {
                // calculate $bias$
                xt::xtensor<double, 3> bias =
                        xt::zeros<double>({n_msk, n_exp, n_srs});

                for (std::size_t m = 0; m < n_msk; m++)
                {
                    // apply the mask
                    // (using NaN workaround until reducers work on masked_view)
                    auto prd_masked = xt::where(xt::view(t_msk, xt::all(), m),
                                                q_prd, NAN);
                    auto obs_masked = xt::where(xt::view(t_msk, xt::all(), m),
                                                q_obs, NAN);

                    // compute variable one sample at a time
                    for (std::size_t e = 0; e < n_exp; e++)
                    {
                        auto prd = xt::view(prd_masked, xt::all(), b_exp[e]);
                        auto obs = xt::view(obs_masked, xt::all(), b_exp[e]);
                        xt::view(bias, m, e) =
                                xt::nansum(prd, -1) / xt::nansum(obs, -1);
                    }
                }

                return bias;
            }
        }
    }
}

#endif //EVALHYD_DETERMINIST_ELEMENTS_HPP
