// Copyright (c) 2023, INRAE.
// Distributed under the terms of the GPL-3 Licence.
// The full licence is in the file LICENCE, distributed with this software.

#ifndef EVALHYD_DETERMINIST_EVALUATOR_HPP
#define EVALHYD_DETERMINIST_EVALUATOR_HPP

#include <vector>

#include <xtl/xoptional.hpp>
#include <xtensor/xexpression.hpp>
#include <xtensor/xtensor.hpp>

#include "elements.hpp"
#include "metrics.hpp"


namespace evalhyd
{
    namespace determinist
    {
        template <class XD2o, class XD2p, class XB3>
        class Evaluator
        {
        private:
            // members for input data
            const XD2o& q_obs;
            const XD2p& q_prd;
            // members for optional input data
            XB3 t_msk;
            const std::vector<xt::xkeep_slice<int>>& b_exp;

            // members for dimensions
            std::size_t n_tim;
            std::size_t n_msk;
            std::size_t n_srs;
            std::size_t n_exp;

            // members for computational elements
            xtl::xoptional<xt::xtensor<double, 4>, bool> mean_obs;
            xtl::xoptional<xt::xtensor<double, 4>, bool> mean_prd;
            xtl::xoptional<xt::xtensor<double, 2>, bool> quad_err;
            xtl::xoptional<xt::xtensor<double, 4>, bool> quad_obs;
            xtl::xoptional<xt::xtensor<double, 4>, bool> quad_prd;
            xtl::xoptional<xt::xtensor<double, 3>, bool> r_pearson;
            xtl::xoptional<xt::xtensor<double, 3>, bool> alpha;
            xtl::xoptional<xt::xtensor<double, 3>, bool> bias;

            // members for evaluation metrics
            xtl::xoptional<xt::xtensor<double, 3>, bool> RMSE;
            xtl::xoptional<xt::xtensor<double, 3>, bool> NSE;
            xtl::xoptional<xt::xtensor<double, 3>, bool> KGE;
            xtl::xoptional<xt::xtensor<double, 3>, bool> KGEPRIME;

            // methods to compute elements
            xt::xtensor<double, 4> get_mean_obs()
            {
                if (!mean_obs.has_value())
                {
                    mean_obs = elements::calc_mean_obs(
                            q_obs, t_msk, b_exp, n_srs, n_msk, n_exp
                    );
                }
                return mean_obs.value();
            };

            xt::xtensor<double, 4> get_mean_prd()
            {
                if (!mean_prd.has_value())
                {
                    mean_prd = elements::calc_mean_prd(
                            q_prd, t_msk, b_exp, n_srs, n_msk, n_exp
                    );
                }
                return mean_prd.value();
            };

            xt::xtensor<double, 2> get_quad_err()
            {
                if (!quad_err.has_value())
                {
                    quad_err = elements::calc_quad_err(
                            q_obs, q_prd
                    );
                }
                return quad_err.value();
            };

            xt::xtensor<double, 4> get_quad_obs()
            {
                if (!quad_obs.has_value())
                {
                    quad_obs = elements::calc_quad_obs(
                            q_obs, get_mean_obs(), t_msk,
                            n_srs, n_tim, n_msk, n_exp
                    );
                }
                return quad_obs.value();
            };

            xt::xtensor<double, 4> get_quad_prd()
            {
                if (!quad_prd.has_value())
                {
                    quad_prd = elements::calc_quad_prd(
                            q_prd, get_mean_prd(), t_msk,
                            n_srs, n_tim, n_msk, n_exp
                    );
                }
                return quad_prd.value();
            };

            xt::xtensor<double, 3> get_r_pearson()
            {
                if (!r_pearson.has_value())
                {
                    r_pearson = elements::calc_r_pearson(
                            q_obs, q_prd, get_mean_obs(), get_mean_prd(),
                            get_quad_obs(), get_quad_prd(), t_msk, b_exp,
                            n_srs, n_msk, n_exp
                    );
                }
                return r_pearson.value();
            };

            xt::xtensor<double, 3> get_alpha()
            {
                if (!alpha.has_value())
                {
                    alpha = elements::calc_alpha(
                            q_obs, q_prd, get_mean_obs(), get_mean_prd(),
                            t_msk, b_exp, n_srs, n_msk, n_exp
                    );
                }
                return alpha.value();
            };

            xt::xtensor<double, 3> get_bias()
            {
                if (!bias.has_value())
                {
                    bias = elements::calc_bias(
                            q_obs, q_prd, t_msk, b_exp, n_srs, n_msk, n_exp
                    );
                }
                return bias.value();
            };

        public:
            // constructor method
            Evaluator(const XD2o& obs,
                      const XD2p& prd,
                      const XB3& msk,
                      const std::vector<xt::xkeep_slice<int>>& exp) :
                    q_obs{obs}, q_prd{prd}, t_msk{msk}, b_exp{exp}
            {
                // initialise a mask if none provided
                // (corresponding to no temporal subset)
                if (msk.size() < 1)
                {
                    t_msk = xt::ones<bool>(
                            {q_prd.shape(0), std::size_t {1}, q_prd.shape(1)}
                    );
                }

                // determine size for recurring dimensions
                n_srs = q_prd.shape(0);
                n_tim = q_prd.shape(1);
                n_msk = msk.shape(1);
                n_exp = b_exp.size();

                // drop time steps where observations or predictions are NaN
                for (std::size_t s = 0; s < n_srs; s++)
                {
                    auto obs_nan = xt::isnan(xt::view(q_obs, 0));
                    auto prd_nan = xt::isnan(xt::view(q_prd, s));

                    auto msk_nan = xt::where(obs_nan || prd_nan)[0];

                    xt::view(t_msk, s, xt::all(), xt::keep(msk_nan)) = false;
                }
            };

            Evaluator(XD2o&& obs,
                      XD2p&& prd,
                      const XB3& msk,
                      const std::vector<xt::xkeep_slice<int>>& exp) :
                    q_obs{obs}, q_prd{prd}, t_msk{msk}, b_exp{exp}
            {
                std::cout << "universal" << std::endl;
                // initialise a mask if none provided
                // (corresponding to no temporal subset)
                if (msk.size() < 1)
                {
                    t_msk = xt::ones<bool>(
                            {q_prd.shape(0), std::size_t {1}, q_prd.shape(1)}
                    );
                }

                // determine size for recurring dimensions
                n_srs = q_prd.shape(0);
                n_tim = q_prd.shape(1);
                n_msk = msk.shape(1);
                n_exp = b_exp.size();

                // drop time steps where observations or predictions are NaN
                for (std::size_t s = 0; s < n_srs; s++)
                {
                    auto obs_nan = xt::isnan(xt::view(q_obs, 0));
                    auto prd_nan = xt::isnan(xt::view(q_prd, s));

                    auto msk_nan = xt::where(obs_nan || prd_nan)[0];

                    xt::view(t_msk, s, xt::all(), xt::keep(msk_nan)) = false;
                }
            };

            // methods to compute metrics
            xt::xtensor<double, 3> get_RMSE()
            {
                if (!RMSE.has_value())
                {
                    RMSE = metrics::calc_RMSE(
                            get_quad_err(), t_msk, b_exp, n_srs, n_msk, n_exp
                    );
                }
                return RMSE.value();
            };

            xt::xtensor<double, 3> get_NSE()
            {
                if (!NSE.has_value())
                {
                    NSE = metrics::calc_NSE(
                            get_quad_err(), get_quad_obs(), t_msk, b_exp,
                            n_srs, n_msk, n_exp
                    );
                }
                return NSE.value();
            };

            xt::xtensor<double, 3> get_KGE()
            {
                if (!KGE.has_value())
                {
                    KGE = metrics::calc_KGE(
                            get_r_pearson(), get_alpha(), get_bias(),
                            n_srs, n_msk, n_exp
                    );
                }
                return KGE.value();
            };

            xt::xtensor<double, 3> get_KGEPRIME()
            {
                if (!KGEPRIME.has_value())
                {
                    KGEPRIME = metrics::calc_KGEPRIME(
                            get_mean_obs(), get_mean_prd(),
                            get_r_pearson(), get_alpha(), get_bias(),
                            n_srs, n_msk, n_exp
                    );
                }
                return KGEPRIME.value();
            };
        };
    }
}

#endif //EVALHYD_DETERMINIST_EVALUATOR_HPP
