// Copyright (c) 2023, INRAE.
// Distributed under the terms of the GPL-3 Licence.
// The full licence is in the file LICENCE, distributed with this software.

#ifndef EVALHYD_DETERMINIST_METRICS_HPP
#define EVALHYD_DETERMINIST_METRICS_HPP

#include <xtensor/xtensor.hpp>
#include <xtensor/xview.hpp>
#include <xtensor/xoperation.hpp>


namespace evalhyd
{
    namespace determinist
    {
        namespace metrics
        {
            /// Compute the root-mean-square error (RMSE).
            ///
            /// \param quad_err
            ///     Quadratic errors between observations and predictions.
            ///     shape: (series, time)
            /// \param t_msk
            ///     Temporal subsets of the whole record.
            ///     shape: (series, subsets, time)
            /// \param b_exp
            ///     Boostrap samples.
            ///     shape: (samples, time slice)
            /// \param n_srs
            ///     Number of prediction series.
            /// \param n_msk
            ///     Number of temporal subsets.
            /// \param n_exp
            ///     Number of bootstrap samples.
            /// \return
            ///     Root-mean-square errors.
            ///     shape: (series, subsets, samples)
            inline xt::xtensor<double, 3> calc_RMSE(
                    const xt::xtensor<double, 2>& quad_err,
                    const xt::xtensor<bool, 3>& t_msk,
                    const std::vector<xt::xkeep_slice<int>>& b_exp,
                    std::size_t n_srs,
                    std::size_t n_msk,
                    std::size_t n_exp
            )
            {
                // compute RMSE
                xt::xtensor<double, 3> RMSE =
                        xt::zeros<double>({n_srs, n_msk, n_exp});

                for (std::size_t m = 0; m < n_msk; m++)
                {
                    // apply the mask
                    // (using NaN workaround until reducers work on masked_view)
                    auto quad_err_masked = xt::where(xt::view(t_msk, xt::all(), m),
                                                     quad_err, NAN);

                    // compute variable one sample at a time
                    for (std::size_t e = 0; e < n_exp; e++)
                    {
                        auto err2 = xt::view(quad_err_masked, xt::all(), b_exp[e]);
                        xt::view(RMSE, xt::all(), m, e) = xt::sqrt(xt::nanmean(err2, -1));
                    }
                }

                return RMSE;
            }

            /// Compute the Nash-Sutcliffe Efficiency (NSE).
            ///
            /// \param quad_err
            ///     Quadratic errors between observations and predictions.
            ///     shape: (series, time)
            /// \param quad_obs
            ///     Quadratic errors between observations and mean observation.
            ///     shape: (subsets, samples, series, time)
            /// \param t_msk
            ///     Temporal subsets of the whole record.
            ///     shape: (series, subsets, time)
            /// \param b_exp
            ///     Boostrap samples.
            ///     shape: (samples, time slice)
            /// \param n_srs
            ///     Number of prediction series.
            /// \param n_msk
            ///     Number of temporal subsets.
            /// \param n_exp
            ///     Number of bootstrap samples.
            /// \return
            ///     Nash-Sutcliffe efficiencies.
            ///     shape: (series, subsets, samples)
            inline xt::xtensor<double, 3> calc_NSE(
                    const xt::xtensor<double, 2>& quad_err,
                    const xt::xtensor<double, 4>& quad_obs,
                    const xt::xtensor<bool, 3>& t_msk,
                    const std::vector<xt::xkeep_slice<int>>& b_exp,
                    std::size_t n_srs,
                    std::size_t n_msk,
                    std::size_t n_exp
            )
            {
                xt::xtensor<double, 3> NSE =
                        xt::zeros<double>({n_srs, n_msk, n_exp});

                for (std::size_t m = 0; m < n_msk; m++)
                {
                    // apply the mask
                    // (using NaN workaround until reducers work on masked_view)
                    auto quad_err_masked = xt::where(xt::view(t_msk, xt::all(), m),
                                                     quad_err, NAN);

                    // compute variable one sample at a time
                    for (std::size_t e = 0; e < n_exp; e++)
                    {
                        // compute squared errors operands
                        auto err2 = xt::view(quad_err_masked, xt::all(), b_exp[e]);
                        xt::xtensor<double, 1> f_num =
                                xt::nansum(err2, -1);
                        auto obs2 = xt::view(quad_obs, m, e, xt::all(), b_exp[e]);
                        xt::xtensor<double, 1> f_den =
                                xt::nansum(obs2, -1);

                        // compute NSE
                        xt::view(NSE, xt::all(), m, e) = 1 - (f_num / f_den);
                    }
                }

                return NSE;
            }

            /// Compute the Kling-Gupta Efficiency (KGE).
            ///
            /// \param r_pearson
            ///     Pearson correlation coefficients.
            ///     shape: (subsets, samples, series)
            /// \param alpha
            ///     Alphas, ratios of standard deviations.
            ///     shape: (subsets, samples, series)
            /// \param bias
            ///     Biases.
            ///     shape: (subsets, samples, series)
            /// \param n_srs
            ///     Number of prediction series.
            /// \param n_msk
            ///     Number of temporal subsets.
            /// \param n_exp
            ///     Number of bootstrap samples.
            /// \return
            ///     Kling-Gupta efficiencies.
            ///     shape: (series, subsets, samples)
            inline xt::xtensor<double, 3> calc_KGE(
                    const xt::xtensor<double, 3>& r_pearson,
                    const xt::xtensor<double, 3>& alpha,
                    const xt::xtensor<double, 3>& bias,
                    std::size_t n_srs,
                    std::size_t n_msk,
                    std::size_t n_exp
            )
            {
                xt::xtensor<double, 3> KGE =
                        xt::zeros<double>({n_srs, n_msk, n_exp});

                for (std::size_t m = 0; m < n_msk; m++)
                {
                    for (std::size_t e = 0; e < n_exp; e++)
                    {
                        // compute KGE
                        xt::view(KGE, xt::all(), m, e) = 1 - xt::sqrt(
                                xt::square(xt::view(r_pearson, m, e) - 1)
                                + xt::square(xt::view(alpha, m, e) - 1)
                                + xt::square(xt::view(bias, m, e) - 1)
                        );
                    }
                }

                return KGE;
            }

            /// Compute the modified Kling-Gupta Efficiency (KGEPRIME).
            ///
            /// \param mean_obs
            ///     Mean observed streamflow.
            ///     shape: (subsets, samples, series, 1)
            /// \param mean_prd
            ///     Mean predicted streamflow.
            ///     shape: (subsets, samples, series, 1)
            /// \param r_pearson
            ///     Pearson correlation coefficients.
            ///     shape: (subsets, samples, series)
            /// \param alpha
            ///     Alphas, ratios of standard deviations.
            ///     shape: (subsets, samples, series)
            /// \param bias
            ///     Biases.
            ///     shape: (subsets, samples, series)
            /// \param n_srs
            ///     Number of prediction series.
            /// \param n_msk
            ///     Number of temporal subsets.
            /// \param n_exp
            ///     Number of bootstrap samples.
            /// \return
            ///     Modified Kling-Gupta efficiencies.
            ///     shape: (series, subsets, samples)
            inline xt::xtensor<double, 3> calc_KGEPRIME(
                    const xt::xtensor<double, 4>& mean_obs,
                    const xt::xtensor<double, 4>& mean_prd,
                    const xt::xtensor<double, 3>& r_pearson,
                    const xt::xtensor<double, 3>& alpha,
                    const xt::xtensor<double, 3>& bias,
                    std::size_t n_srs,
                    std::size_t n_msk,
                    std::size_t n_exp
            )
            {
                xt::xtensor<double, 3> KGEPRIME =
                        xt::zeros<double>({n_srs, n_msk, n_exp});

                for (std::size_t m = 0; m < n_msk; m++)
                {
                    for (std::size_t e = 0; e < n_exp; e++)
                    {
                        // calculate error in spread of flow $gamma$
                        auto gamma = xt::view(alpha, m, e)
                                     * (xt::view(mean_obs, m, e, xt::all(), 0)
                                        / xt::view(mean_prd, m, e, xt::all(), 0));

                        // compute KGEPRIME
                        xt::view(KGEPRIME, xt::all(), m, e) =  1 - xt::sqrt(
                                xt::square(xt::view(r_pearson, m, e) - 1)
                                + xt::square(gamma - 1)
                                + xt::square(xt::view(bias, m, e) - 1)
                        );
                    }
                }

                return KGEPRIME;
            }
        }
    }
}

#endif //EVALHYD_DETERMINIST_METRICS_HPP
