// Copyright (c) 2023, INRAE.
// Distributed under the terms of the GPL-3 Licence.
// The full licence is in the file LICENCE, distributed with this software.

#ifndef EVALHYD_PROBABILIST_BRIER_HPP
#define EVALHYD_PROBABILIST_BRIER_HPP

#include <xtensor/xtensor.hpp>
#include <xtensor/xview.hpp>
#include <xtensor/xmasked_view.hpp>


// NOTE ------------------------------------------------------------------------
// All equations in metrics below are following notations from
// "Wilks, D. S. (2011). Statistical methods in the atmospheric sciences.
// Amsterdam; Boston: Elsevier Academic Press. ISBN: 9780123850225".
// In particular, pp. 302-303, 332-333.
// -----------------------------------------------------------------------------

namespace evalhyd
{
    namespace probabilist
    {
        namespace elements
        {
            /// Determine observed realisation of threshold(s) exceedance.
            ///
            /// \param q_obs
            ///     Streamflow observations.
            ///     shape: (sites, time)
            /// \param q_thr
            ///     Streamflow exceedance threshold(s).
            ///     shape: (sites, thresholds)
            /// \return
            ///     Event observed outcome.
            ///     shape: (sites, thresholds, time)
            template<class XD2>
            inline xt::xtensor<double, 3> calc_o_k(
                    const XD2& q_obs,
                    const XD2& q_thr,
                    bool is_high_flow_event
            )
            {
                if (is_high_flow_event)
                {
                    // observations above threshold(s)
                    return xt::view(q_obs, xt::all(), xt::newaxis(), xt::all())
                           >= xt::view(q_thr, xt::all(), xt::all(), xt::newaxis());
                }
                else
                {
                    // observations below threshold(s)
                    return xt::view(q_obs, xt::all(), xt::newaxis(), xt::all())
                           < xt::view(q_thr, xt::all(), xt::all(), xt::newaxis());
                }

            }

            /// Determine mean observed realisation of threshold(s) exceedance.
            ///
            /// \param o_k
            ///     Event observed outcome.
            ///     shape: (sites, thresholds, time)
            /// \param t_msk
            ///     Temporal subsets of the whole record.
            ///     shape: (sites, lead times, subsets, time)
            /// \param b_exp
            ///     Boostrap samples.
            ///     shape: (samples, time slice)
            /// \param n_sit
            ///     Number of sites.
            /// \param n_ldt
            ///     Number of lead times.
            /// \param n_thr
            ///     Number of thresholds.
            /// \param n_msk
            ///     Number of temporal subsets.
            /// \param n_exp
            ///     Number of bootstrap samples.
            /// \return
            ///     Mean event observed outcome.
            ///     shape: (sites, lead times, subsets, samples, thresholds)
            inline xt::xtensor<double, 5> calc_bar_o(
                    const xt::xtensor<double, 3>& o_k,
                    const xt::xtensor<bool, 4>& t_msk,
                    const std::vector<xt::xkeep_slice<int>>& b_exp,
                    std::size_t n_sit,
                    std::size_t n_ldt,
                    std::size_t n_thr,
                    std::size_t n_msk,
                    std::size_t n_exp
            )
            {
                // initialise output variable
                xt::xtensor<double, 5> bar_o =
                        xt::zeros<double>({n_sit, n_ldt, n_msk, n_exp, n_thr});

                // apply the mask
                // (using NaN workaround until reducers work on masked_view)
                auto o_k_masked = xt::where(
                        xt::view(t_msk, xt::all(), xt::all(),
                                 xt::all(), xt::newaxis(), xt::all()),
                        xt::view(o_k, xt::all(), xt::newaxis(),
                                 xt::newaxis(), xt::all(), xt::all()),
                        NAN
                );

                // compute variable one sample at a time
                for (std::size_t e = 0; e < n_exp; e++)
                {
                    // apply the bootstrap sampling
                    auto o_k_masked_sampled =
                            xt::view(o_k_masked, xt::all(), xt::all(),
                                     xt::all(), xt::all(), b_exp[e]);

                    // compute mean "climatology" relative frequency of the event
                    // $\bar{o} = \frac{1}{n} \sum_{k=1}^{n} o_k$
                    xt::view(bar_o, xt::all(), xt::all(), xt::all(), e, xt::all()) =
                            xt::nanmean(o_k_masked_sampled, -1);
                }

                return bar_o;
            }

            /// Determine number of forecast members exceeding threshold(s)
            ///
            /// \param q_prd
            ///     Streamflow predictions.
            ///     shape: (sites, lead times, members, time)
            /// \param q_thr
            ///     Streamflow exceedance threshold(s).
            ///     shape: (sites, thresholds)
            /// \return
            ///     Number of forecast members exceeding threshold(s).
            ///     shape: (sites, lead times, thresholds, time)
            template<class XD4, class XD2>
            inline xt::xtensor<double, 4> calc_sum_f_k(
                    const XD4& q_prd,
                    const XD2& q_thr,
                    bool is_high_flow_event
            )
            {
                if (is_high_flow_event)
                {
                    // determine if members are above threshold(s)
                    auto f_k = xt::view(q_prd, xt::all(), xt::all(),
                                        xt::newaxis(), xt::all(), xt::all())
                               >= xt::view(q_thr, xt::all(), xt::newaxis(),
                                           xt::all(), xt::newaxis(), xt::newaxis());

                    // calculate how many members are above threshold(s)
                    return xt::sum(f_k, 3);
                }
                else
                {
                    // determine if members are below threshold(s)
                    auto f_k = xt::view(q_prd, xt::all(), xt::all(),
                                        xt::newaxis(), xt::all(), xt::all())
                               < xt::view(q_thr, xt::all(), xt::newaxis(),
                                           xt::all(), xt::newaxis(), xt::newaxis());

                    // calculate how many members are below threshold(s)
                    return xt::sum(f_k, 3);
                }
            }

            /// Determine forecast probability of threshold(s) exceedance to occur.
            ///
            /// \param sum_f_k
            ///     Number of forecast members exceeding threshold(s).
            ///     shape: (sites, lead times, thresholds, time)
            /// \param n_mbr
            ///     Number of ensemble members.
            /// \return
            ///     Event probability forecast.
            ///     shape: (sites, lead times, thresholds, time)
            inline xt::xtensor<double, 4> calc_y_k(
                    const xt::xtensor<double, 4>& sum_f_k,
                    std::size_t n_mbr
            )
            {
                // determine probability of threshold(s) exceedance
                // /!\ probability calculation dividing by n (the number of
                //     members), not n+1 (the number of ranks) like in other metrics
                return sum_f_k / n_mbr;
            }
        }

        namespace intermediate
        {
            /// Compute the Brier score for each time step.
            ///
            /// \param o_k
            ///     Observed event outcome.
            ///     shape: (sites, thresholds, time)
            /// \param y_k
            ///     Event probability forecast.
            ///     shape: (sites, lead times, thresholds, time)
            /// \return
            ///     Brier score for each threshold for each time step.
            ///     shape: (sites, lead times, thresholds, time)
            inline xt::xtensor<double, 4> calc_bs(
                    const xt::xtensor<double, 3>& o_k,
                    const xt::xtensor<double, 4>& y_k
            )
            {
                // return computed Brier score(s)
                // $bs = (o_k - y_k)^2$
                return xt::square(
                        xt::view(o_k, xt::all(), xt::newaxis(),
                                 xt::all(), xt::all())
                        - y_k
                );
            }
        }

        namespace metrics
        {
            /// Compute the Brier score (BS).
            ///
            /// \param bs
            ///     Brier score for each threshold for each time step.
            ///     shape: (sites, lead times, thresholds, time)
            /// \param q_thr
            ///     Streamflow exceedance threshold(s).
            ///     shape: (sites, thresholds)
            /// \param t_msk
            ///     Temporal subsets of the whole record.
            ///     shape: (sites, lead times, subsets, time)
            /// \param b_exp
            ///     Boostrap samples.
            ///     shape: (samples, time slice)
            /// \param n_sit
            ///     Number of sites.
            /// \param n_ldt
            ///     Number of lead times.
            /// \param n_thr
            ///     Number of thresholds.
            /// \param n_msk
            ///     Number of temporal subsets.
            /// \param n_exp
            ///     Number of bootstrap samples.
            /// \return
            ///     Brier score for each subset and for each threshold.
            ///     shape: (sites, lead times, subsets, samples, thresholds)
            template <class XD2>
            inline xt::xtensor<double, 5> calc_BS(
                    const xt::xtensor<double, 4>& bs,
                    const XD2& q_thr,
                    const xt::xtensor<bool, 4>& t_msk,
                    const std::vector<xt::xkeep_slice<int>>& b_exp,
                    std::size_t n_sit,
                    std::size_t n_ldt,
                    std::size_t n_thr,
                    std::size_t n_msk,
                    std::size_t n_exp
            )
            {
                // initialise output variable
                xt::xtensor<double, 5> BS =
                        xt::zeros<double>({n_sit, n_ldt, n_msk, n_exp, n_thr});

                // compute variable one mask at a time to minimise memory imprint
                for (std::size_t m = 0; m < n_msk; m++)
                {
                    // apply the mask
                    // (using NaN workaround until reducers work on masked_view)
                    auto bs_masked = xt::where(
                            xt::view(t_msk, xt::all(), xt::all(), m,
                                     xt::newaxis(), xt::all()),
                            bs,
                            NAN
                    );

                    // compute variable one sample at a time
                    for (std::size_t e = 0; e < n_exp; e++)
                    {
                        // apply the bootstrap sampling
                        auto bs_masked_sampled =
                                xt::view(bs_masked, xt::all(), xt::all(),
                                         xt::all(), b_exp[e]);

                        // calculate the mean over the time steps
                        // $BS = \frac{1}{n} \sum_{k=1}^{n} (o_k - y_k)^2$
                        xt::view(BS, xt::all(), xt::all(), m, e, xt::all()) =
                                xt::nanmean(bs_masked_sampled, -1);
                    }
                }

                // assign NaN where thresholds were not provided (i.e. set as NaN)
                xt::masked_view(
                        BS,
                        xt::isnan(xt::view(q_thr, xt::all(), xt::newaxis(),
                                           xt::newaxis(), xt::newaxis(),
                                           xt::all()))
                ) = NAN;

                return BS;
            }

            /// Compute the calibration-refinement decomposition of the Brier score
            /// into reliability, resolution, and uncertainty.
            ///
            /// BS = reliability - resolution + uncertainty
            ///
            /// \param q_thr
            ///     Streamflow exceedance threshold(s).
            ///     shape: (sites, thresholds)
            /// \param o_k
            ///     Observed event outcome.
            ///     shape: (sites, thresholds, time)
            /// \param y_k
            ///     Event probability forecast.
            ///     shape: (sites, lead times, thresholds, time)
            /// \param bar_o
            ///     Mean event observed outcome.
            ///     shape: (sites, lead times, subsets, samples, thresholds)
            /// \param t_msk
            ///     Temporal subsets of the whole record.
            ///     shape: (sites, lead times, subsets, time)
            /// \param b_exp
            ///     Boostrap samples.
            ///     shape: (samples, time slice)
            /// \param n_sit
            ///     Number of sites.
            /// \param n_ldt
            ///     Number of lead times.
            /// \param n_thr
            ///     Number of thresholds.
            /// \param n_mbr
            ///     Number of ensemble members.
            /// \param n_msk
            ///     Number of temporal subsets.
            /// \param n_exp
            ///     Number of bootstrap samples.
            /// \return
            ///     Brier score components (reliability, resolution, uncertainty)
            ///     for each subset and for each threshold.
            ///     shape: (sites, lead times, subsets, samples, thresholds, components)
            template <class XD2>
            inline xt::xtensor<double, 6> calc_BS_CRD(
                    const XD2& q_thr,
                    const xt::xtensor<double, 3>& o_k,
                    const xt::xtensor<double, 4>& y_k,
                    const xt::xtensor<double, 5>& bar_o,
                    const xt::xtensor<bool, 4>& t_msk,
                    const std::vector<xt::xkeep_slice<int>>& b_exp,
                    std::size_t n_sit,
                    std::size_t n_ldt,
                    std::size_t n_thr,
                    std::size_t n_mbr,
                    std::size_t n_msk,
                    std::size_t n_exp
            )
            {
                // declare internal variables
                // shape: (sites, lead times, bins, thresholds, time)
                xt::xtensor<double, 5> msk_bins;
                // shape: (sites, lead times, bins, thresholds)
                xt::xtensor<double, 4> N_i, bar_o_i;
                // shape: (bins,)
                xt::xtensor<double, 1> y_i;

                // compute range of forecast values $y_i$
                y_i = xt::arange<double>(double(n_mbr + 1)) / n_mbr;

                // initialise output variable
                xt::xtensor<double, 6> BS_CRD =
                        xt::zeros<double>({n_sit, n_ldt, n_msk, n_exp, n_thr,
                                           std::size_t {3}});

                // compute variable one mask at a time to minimise memory imprint
                for (std::size_t m = 0; m < n_msk; m++)
                {
                    // apply the mask
                    // (using NaN workaround until reducers work on masked_view)
                    auto o_k_masked = xt::where(
                            xt::view(t_msk, xt::all(), xt::all(),
                                     m, xt::newaxis(), xt::all()),
                            xt::view(o_k, xt::all(), xt::newaxis(),
                                     xt::all(), xt::all()),
                            NAN
                    );
                    auto y_k_masked = xt::where(
                            xt::view(t_msk, xt::all(), xt::all(),
                                     m, xt::newaxis(), xt::all()),
                            y_k,
                            NAN
                    );

                    // compute variable one sample at a time
                    for (std::size_t e = 0; e < n_exp; e++)
                    {
                        // apply the bootstrap sampling
                        auto o_k_masked_sampled =
                                xt::view(o_k_masked, xt::all(), xt::all(),
                                         xt::all(), b_exp[e]);
                        auto y_k_masked_sampled =
                                xt::view(y_k_masked, xt::all(), xt::all(),
                                         xt::all(), b_exp[e]);
                        auto t_msk_sampled =
                                xt::view(t_msk, xt::all(), xt::all(), m,
                                         xt::newaxis(), b_exp[e]);
                        auto bar_o_sampled =
                                xt::view(bar_o, xt::all(), xt::all(),
                                         xt::all(), e, xt::all());

                        // calculate length of subset
                        auto l = xt::sum(t_msk_sampled, -1);  // (sit, ldt, 1)

                        // compute mask to subsample time steps belonging to same forecast bin
                        // (where bins are defined as the range of forecast values)
                        msk_bins = xt::equal(
                                // force evaluation to avoid segfault
                                xt::view(xt::eval(y_k_masked_sampled),
                                         xt::all(), xt::all(), xt::newaxis(),
                                         xt::all(), xt::all()),
                                xt::view(y_i,
                                         xt::newaxis(), xt::newaxis(), xt::all(),
                                         xt::newaxis(), xt::newaxis())
                        );

                        // compute number of forecasts in each forecast bin $N_i$
                        N_i = xt::nansum(msk_bins, -1);

                        // compute subsample relative frequency
                        // $\bar{o_i} = \frac{1}{N_i} \sum_{k \in N_i} o_k$
                        bar_o_i = xt::where(
                                N_i > 0,
                                xt::nansum(
                                        xt::where(
                                                msk_bins,
                                                xt::view(o_k_masked_sampled,
                                                         xt::all(), xt::all(),
                                                         xt::newaxis(),
                                                         xt::all(), xt::all()),
                                                0.
                                        ),
                                        -1
                                ) / N_i,
                                0.
                        );

                        // calculate reliability =
                        // $\frac{1}{n} \sum_{i=1}^{I} N_i (y_i - \bar{o_i})^2$
                        xt::view(BS_CRD, xt::all(), xt::all(), m, e, xt::all(), 0) =
                                xt::nansum(
                                        xt::square(
                                                xt::view(y_i, xt::newaxis(),
                                                         xt::newaxis(), xt::all(),
                                                         xt::newaxis())
                                                - bar_o_i
                                        ) * N_i,
                                        2
                                ) / l;

                        // calculate resolution =
                        // $\frac{1}{n} \sum_{i=1}^{I} N_i (\bar{o_i} - \bar{o})^2$
                        xt::view(BS_CRD, xt::all(), xt::all(), m, e, xt::all(), 1) =
                                xt::nansum(
                                        xt::square(
                                                bar_o_i
                                                - xt::view(bar_o_sampled,
                                                           xt::all(), xt::all(),
                                                           m, xt::newaxis(),
                                                           xt::all())
                                        ) * N_i,
                                        2
                                ) / l;

                        // calculate uncertainty = $\bar{o} (1 - \bar{o})$
                        xt::view(BS_CRD, xt::all(), xt::all(), m, e, xt::all(), 2) =
                                xt::view(bar_o_sampled, xt::all(), xt::all(), m)
                                * (1 - xt::view(bar_o_sampled, xt::all(), xt::all(), m));
                    }
                }

                // assign NaN where thresholds were not provided (i.e. set as NaN)
                xt::masked_view(
                        BS_CRD,
                        xt::isnan(xt::view(q_thr, xt::all(), xt::newaxis(),
                                           xt::newaxis(), xt::newaxis(),
                                           xt::all(), xt::newaxis()))
                ) = NAN;

                return BS_CRD;
            }

            /// Compute the likelihood-base rate decomposition of the Brier score
            /// into type 2 bias, discrimination, and sharpness (a.k.a. refinement).
            ///
            /// BS = type 2 bias - discrimination + sharpness
            ///
            /// \param q_thr
            ///     Streamflow exceedance threshold(s).
            ///     shape: (sites, thresholds)
            /// \param o_k
            ///     Observed event outcome.
            ///     shape: (sites, thresholds, time)
            /// \param y_k
            ///     Event probability forecast.
            ///     shape: (sites, lead times, thresholds, time)
            /// \param t_msk
            ///     Temporal subsets of the whole record.
            ///     shape: (sites, lead times, subsets, time)
            /// \param b_exp
            ///     Boostrap samples.
            ///     shape: (samples, time slice)
            /// \param n_sit
            ///     Number of sites.
            /// \param n_ldt
            ///     Number of lead times.
            /// \param n_thr
            ///     Number of thresholds.
            /// \param n_msk
            ///     Number of temporal subsets.
            /// \param n_exp
            ///     Number of bootstrap samples.
            /// \return
            ///     Brier score components (type 2 bias, discrimination, sharpness)
            ///     for each subset and for each threshold.
            ///     shape: (sites, lead times, subsets, samples, thresholds, components)
            template <class XD2>
            inline xt::xtensor<double, 6> calc_BS_LBD(
                    const XD2& q_thr,
                    const xt::xtensor<double, 3>& o_k,
                    const xt::xtensor<double, 4>& y_k,
                    const xt::xtensor<bool, 4>& t_msk,
                    const std::vector<xt::xkeep_slice<int>>& b_exp,
                    std::size_t n_sit,
                    std::size_t n_ldt,
                    std::size_t n_thr,
                    std::size_t n_msk,
                    std::size_t n_exp
            )
            {
                // declare internal variables
                // shape: (sites, lead times, bins, thresholds, time)
                xt::xtensor<double, 5> msk_bins;
                // shape: (sites, lead times, thresholds)
                xt::xtensor<double, 3> bar_y;
                // shape: (sites, lead times, bins, thresholds)
                xt::xtensor<double, 4> M_j, bar_y_j;
                // shape: (bins,)
                xt::xtensor<double, 1> o_j;

                // set the range of observed values $o_j$
                o_j = {0., 1.};

                // declare and initialise output variable
                xt::xtensor<double, 6> BS_LBD =
                        xt::zeros<double>({n_sit, n_ldt, n_msk, n_exp, n_thr,
                                           std::size_t {3}});

                // compute variable one mask at a time to minimise memory imprint
                for (std::size_t m = 0; m < n_msk; m++)
                {
                    // apply the mask
                    // (using NaN workaround until reducers work on masked_view)
                    auto o_k_masked = xt::where(
                            xt::view(t_msk, xt::all(), xt::all(),
                                     m, xt::newaxis(), xt::all()),
                            xt::view(o_k, xt::all(), xt::newaxis(),
                                     xt::all(), xt::all()),
                            NAN
                    );
                    auto y_k_masked = xt::where(
                            xt::view(t_msk, xt::all(), xt::all(),
                                     m, xt::newaxis(), xt::all()),
                            y_k,
                            NAN
                    );

                    // compute variable one sample at a time
                    for (std::size_t e = 0; e < n_exp; e++)
                    {
                        // apply the bootstrap sampling
                        auto o_k_masked_sampled =
                                xt::view(o_k_masked, xt::all(), xt::all(),
                                         xt::all(), b_exp[e]);
                        auto y_k_masked_sampled =
                                xt::view(y_k_masked, xt::all(), xt::all(),
                                         xt::all(), b_exp[e]);
                        auto t_msk_sampled =
                                xt::view(t_msk, xt::all(), xt::all(), m,
                                         xt::newaxis(), b_exp[e]);

                        // calculate length of subset
                        auto l = xt::sum(t_msk_sampled, -1);

                        // compute mask to subsample time steps belonging to same observation bin
                        // (where bins are defined as the range of forecast values)
                        msk_bins = xt::equal(
                                // force evaluation to avoid segfault
                                xt::view(xt::eval(o_k_masked_sampled),
                                         xt::all(), xt::all(), xt::newaxis(),
                                         xt::all(), xt::all()),
                                xt::view(o_j,
                                         xt::newaxis(), xt::newaxis(), xt::all(),
                                         xt::newaxis(), xt::newaxis())
                        );

                        // compute number of observations in each observation bin $M_j$
                        M_j = xt::nansum(msk_bins, -1);

                        // compute subsample relative frequency
                        // $\bar{y_j} = \frac{1}{M_j} \sum_{k \in M_j} y_k$
                        bar_y_j = xt::where(
                                M_j > 0,
                                xt::nansum(
                                        xt::where(
                                                msk_bins,
                                                xt::view(y_k_masked_sampled,
                                                         xt::all(), xt::all(),
                                                         xt::newaxis(),
                                                         xt::all(), xt::all()),
                                                0.
                                        ),
                                        -1
                                ) / M_j,
                                0.
                        );

                        // compute mean "climatology" forecast probability
                        // $\bar{y} = \frac{1}{n} \sum_{k=1}^{n} y_k$
                        bar_y = xt::nanmean(y_k_masked_sampled, -1);

                        // calculate type 2 bias =
                        // $\frac{1}{n} \sum_{j=1}^{J} M_j (o_j - \bar{y_j})^2$
                        xt::view(BS_LBD, xt::all(), xt::all(), m, e, xt::all(), 0) =
                                xt::nansum(
                                        xt::square(
                                                xt::view(o_j, xt::newaxis(),
                                                         xt::newaxis(), xt::all(),
                                                         xt::newaxis())
                                                - bar_y_j
                                        ) * M_j,
                                        2
                                ) / l;

                        // calculate discrimination =
                        // $\frac{1}{n} \sum_{j=1}^{J} M_j (\bar{y_j} - \bar{y})^2$
                        xt::view(BS_LBD, xt::all(), xt::all(), m, e, xt::all(), 1) =
                                xt::nansum(
                                        xt::square(
                                                bar_y_j
                                                - xt::view(bar_y,
                                                           xt::all(), xt::all(),
                                                           xt::newaxis(),
                                                           xt::all())
                                        ) * M_j,
                                        2
                                ) / l;

                        // calculate sharpness =
                        // $\frac{1}{n} \sum_{k=1}^{n} (\bar{y_k} - \bar{y})^2$
                        xt::view(BS_LBD, xt::all(), xt::all(), m, e, xt::all(), 2) =
                                xt::nansum(
                                        xt::square(
                                                y_k_masked_sampled
                                                - xt::view(bar_y, xt::all(),
                                                           xt::all(), xt::all(),
                                                           xt::newaxis())
                                        ),
                                        -1
                                ) / l;
                    }

                }

                // assign NaN where thresholds were not provided (i.e. set as NaN)
                xt::masked_view(
                        BS_LBD,
                        xt::isnan(xt::view(q_thr, xt::all(), xt::newaxis(),
                                           xt::newaxis(), xt::newaxis(),
                                           xt::all(), xt::newaxis()))
                ) = NAN;

                return BS_LBD;
            }

            /// Compute the Brier skill score (BSS).
            ///
            /// \param bs
            ///     Brier score for each threshold for each time step.
            ///     shape: (sites, lead times, thresholds, time)
            /// \param q_thr
            ///     Streamflow exceedance threshold(s).
            ///     shape: (sites, thresholds)
            /// \param o_k
            ///     Observed event outcome.
            ///     shape: (sites, thresholds, time)
            /// \param bar_o
            ///     Mean event observed outcome.
            ///     shape: (sites, lead times, subsets, samples, thresholds)
            /// \param t_msk
            ///     Temporal subsets of the whole record.
            ///     shape: (sites, lead times, subsets, time)
            /// \param b_exp
            ///     Boostrap samples.
            ///     shape: (samples, time slice)
            /// \param n_sit
            ///     Number of sites.
            /// \param n_ldt
            ///     Number of lead times.
            /// \param n_thr
            ///     Number of thresholds.
            /// \param n_msk
            ///     Number of temporal subsets.
            /// \param n_exp
            ///     Number of bootstrap samples.
            /// \return
            ///     Brier skill score for each subset and for each threshold.
            ///     shape: (sites, lead times, subsets, samples, thresholds)
            template <class XD2>
            inline xt::xtensor<double, 5> calc_BSS(
                    const xt::xtensor<double, 4>& bs,
                    const XD2& q_thr,
                    const xt::xtensor<double, 3>& o_k,
                    const xt::xtensor<double, 5>& bar_o,
                    const xt::xtensor<bool, 4>& t_msk,
                    const std::vector<xt::xkeep_slice<int>>& b_exp,
                    std::size_t n_sit,
                    std::size_t n_ldt,
                    std::size_t n_thr,
                    std::size_t n_msk,
                    std::size_t n_exp
            )
            {
                // declare and initialise output variable
                xt::xtensor<double, 5> BSS =
                        xt::zeros<double>({n_sit, n_ldt, n_msk, n_exp, n_thr});

                // compute variable one mask at a time to minimise memory imprint
                for (std::size_t m = 0; m < n_msk; m++)
                {
                    // apply the mask
                    // (using NaN workaround until reducers work on masked_view)
                    auto o_k_masked = xt::where(
                            xt::view(t_msk, xt::all(), xt::all(),
                                     m, xt::newaxis(), xt::all()),
                            xt::view(o_k, xt::all(), xt::newaxis(),
                                     xt::all(), xt::all()),
                            NAN
                    );
                    auto bs_masked = xt::where(
                            xt::view(t_msk, xt::all(), xt::all(), m,
                                     xt::newaxis(), xt::all()),
                            bs,
                            NAN
                    );

                    // compute variable one sample at a time
                    for (std::size_t e = 0; e < n_exp; e++)
                    {
                        // apply the bootstrap sampling
                        auto o_k_masked_sampled =
                                xt::view(o_k_masked, xt::all(), xt::all(),
                                         xt::all(), b_exp[e]);
                        auto bs_masked_sampled =
                                xt::view(bs_masked, xt::all(), xt::all(),
                                         xt::all(), b_exp[e]);
                        auto bar_o_sampled =
                                xt::view(bar_o, xt::all(), xt::all(),
                                         xt::all(), e, xt::all());

                        // calculate reference Brier score(s)
                        // $bs_{ref} = \frac{1}{n} \sum_{k=1}^{n} (o_k - \bar{o})^2$
                        xt::xtensor<double, 4> bs_ref =
                                xt::nanmean(
                                        xt::square(
                                                o_k_masked_sampled -
                                                xt::view(
                                                        bar_o_sampled, xt::all(),
                                                        xt::all(), m, xt::all(),
                                                        xt::newaxis()
                                                )
                                        ),
                                        -1,
                                        xt::keep_dims
                                );

                        // compute Brier skill score(s)
                        // $BSS = \frac{1}{n} \sum_{k=1}^{n} 1 - \frac{bs}{bs_{ref}}
                        xt::view(BSS, xt::all(), xt::all(), m, e, xt::all()) =
                                xt::nanmean(
                                        xt::where(
                                                xt::equal(bs_ref, 0),
                                                0,
                                                1 - (bs_masked_sampled / bs_ref)
                                        ),
                                        -1
                                );
                    }
                }

                // assign NaN where thresholds were not provided (i.e. set as NaN)
                xt::masked_view(
                        BSS,
                        xt::isnan(xt::view(q_thr, xt::all(), xt::newaxis(),
                                           xt::newaxis(), xt::newaxis(),
                                           xt::all()))
                ) = NAN;

                return BSS;
            }
        }
    }
}

#endif //EVALHYD_PROBABILIST_BRIER_HPP
